
from typing import NewType

ExtensionName        = NewType('ExtensionName', str)
FormatName           = NewType('FormatName', str)
FileSuffix           = NewType('FileSuffix', str)
ExtensionDescription = NewType('ExtensionDescription', str)

Author  = NewType('Author', str)
Version = NewType('Version', str)
